# -*- coding: utf-8 -*-
"""
requirements: pywebview>=3.7.2
"""

import os, webview
from noderedpy import (
    RED, REDBuilder, Auth, Node,
    InputProperty, ListProperty, DictProperty
)
from noderedpy.decorator import register

__dirname = os.path.dirname(os.path.realpath(__file__))
if __name__ == "__main__":
    class TotalApp:
        count = 0

        @register("test", properties = [
            InputProperty("test_prop", "1234"),
            ListProperty("list_prop", [ 1, 2, 3, 4 ]),
            DictProperty("dict_prop", { "a": 1 })
        ])
        def test(node:Node, props:dict, msg:dict) -> dict:
            TotalApp.count += 1

            print(props)
            print(msg)

            msg["payload"] = TotalApp.count

            return msg
        
    # get RED object from builder
    red = REDBuilder()\
        .set_user_dir(os.path.join(__dirname, ".node-red-pywebview"))\
        .set_node_red_dir(os.path.join(__dirname, "node_red_dir"))\
        .set_admin_root("/node-red")\
        .set_port(1880)\
        .set_remote_access(False).build()

    # # init RED object directly
    # red = RED(
    #     os.path.join(__dirname, ".node-red-pywebview"),
    #     os.path.join(__dirname, "node_red_dir"),
    #     "/node-red",
    #     port = 1880,
    #     editor_theme = {
    #         "projects": {
    #             "enabled": False
    #         }
    #     }
    # )

    # set editor theme
    red.editor_theme.palette.editable = False
    red.editor_theme.projects.enabled = False

    # add auths
    red.node_auths.append(
        Auth(username = "node-red-py", password = "p@ssword")
    )

    webview.initialize()
    win = webview.create_window("Node-RED.py pywebview")
    win.events.closing += lambda: red.stop()

    webview.start(lambda: red.start(callback = lambda: win.load_url(f"http://127.0.0.1:{red.port}{red.admin_root}"), start_browser = False), debug = True)



# -*- coding: utf-8 -*-
from dataclasses import dataclass, field
from typing import Literal, List


@dataclass
class Auth:
    username:str
    password:str
    permissions:Literal["*", "read"] = field(default = "*")

    def to_dict(self) -> dict:
        return {
            "username": self.username,
            "password": self.password,
            "permissions": self.permissions
        }

class AuthCollection:
    def __init__(self):
        self.__auths:List[Auth] = []

    def append(self, auth:Auth):
        if not auth in self.__auths:
            self.__auths.append(auth)

    def remove(self, auth:Auth):
        if auth in self.__auths:
            self.__auths.remove(auth)

    def to_list(self) -> List[dict]:
        return [
            auth.to_dict()
            for auth in self.__auths
            if auth.username is not None and auth.password is not None
        ]



# -*- coding: utf-8 -*-
import os, gc, traceback
from types import MethodType


class Route:
    def __init__(self, url:str, method:str, target:MethodType):
        # check url is valid
        if not url.startswith("/"):
            raise ValueError("url must starts with `/`!")

        self.url, self.method = url, method
        self.__target = target

    def run(self, route_data:dict) -> dict:
        gc.enable()

        print(f"\n{self.method} | {self.url} entered\n=============================================")
        try:
            data = self.__target(route_data)

            print("======================================= ended\n")
            gc.collect()

            return { "state": "success", "data": data }
        except:
            return { "state": "fail", "message": traceback.format_exc() }
        
    def to_dict(self) -> dict:
        return {
            "url": self.url,
            "method": self.method
        }

class StaticRoute(Route):
    def __init__(self, url:str, path:os.PathLike):
        super().__init__(url, "static", None)
        self.path = path

    def to_dict(self) -> dict:
        return {
            "url": self.url,
            "method": "static",
            "path": self.path
        }



# -*- coding: utf-8 -*-
from typing import List, Literal
from types import MethodType
from .__nodered__.red import RED
from .__nodered__.node import Node
from .__property__ import Property
from .__nodered__.route import Route


def register(name:str, category:str = "nodered_py", version:str = "1.0.0", description:str = "", author:str = "nodered.py", keywords:List[str] = [], icon:str = "function.png", properties:List[Property] = []) -> MethodType:
    """
    Decorator to register Node function

    Parameters
    ----------
    name: str, required
        name of Node to register
    category: str, default nodered_py
        category of Node
    version: str, default 1.0.0
        version of Node
    description: str, default ""
        description of Node
    author: str, default nodered.py
        author of Node
    keywords: List[str], default []
        extra keywords of Node
    icon: str, default function.png
        icon of Node(html)
    properties: List[noderedpy._property.Property]
        propertis of Node
    """
    def decorator(node_func:MethodType):
        RED.registered_nodes.append(
            Node(
                name, category,
                version, description, author, keywords,
                icon, properties, node_func
            )
        )

        return node_func
    
    return decorator

def route(url:str, method:Literal["get", "post"]) -> MethodType:
    """
    Decorator to register route to Node-RED

    Parameters
    ----------
    url: str, required
        url of route point
    method: str, required
        method of route point
        options: get, post
    """
    def decorator(route_func:MethodType):
        RED.registered_routes.append(
            Route(url, method, route_func)
        )

        return route_func
    
    return decorator



# -*- coding: utf-8 -*-
import os
from noderedpy import (
    REDBuilder, RED, Auth, Node,
    InputProperty, ListProperty, DictProperty, CodeProperty,
    SpinnerProperty, CheckBoxProperty, ComboBoxProperty, PropertyDivider
)
from noderedpy.decorator import register, route

__dirname = os.path.dirname(os.path.realpath(__file__))
if __name__ == "__main__":
    class TotalApp:
        count = 0

        @register("test", properties = [
            InputProperty("test_prop", "1234"),
            ListProperty("list_prop", [ 1, 2, 3, 4 ]),
            DictProperty("dict_prop", { "a": 1 })
        ])
        def test(node:Node, props:dict, msg:dict) -> dict:
            TotalApp.count += 1

            print(props)
            print(msg)

            msg["payload"] = TotalApp.count

            return msg
        
        @register(
            "property-test", author = "oyajiDev", version = "0.1.0", description = "test for multi properties",
            icon = "database.png",
            properties = [
                InputProperty("input_prop", "input property"),
                ListProperty("list_prop", [ "list", "property" ], 150),
                DictProperty("dict_prop", { "dict": "property" }, 100),
                SpinnerProperty("spinner_prop", 1),
                CheckBoxProperty("chkbox_prop", True),
                PropertyDivider(),
                ComboBoxProperty("cbox_prop", [ "combobox", "property" ]),
                CodeProperty("code_prop", "print(1234)", "python")
            ]
        )
        def property_test(node:Node, props:dict, msg:dict) -> dict:
            node.status("grey", "ring", "status test")
            node.log("123", "456", "789")
            node.warn("123", "456", "789")
            node.error("123", "456", "789")

            print(props, msg)
            return msg
        
        def test2(self, node:Node, props:dict, msg:dict) -> dict:
            print(self.count, msg["payload"])

            return msg

    # get RED object from builder
    red = REDBuilder()\
        .set_user_dir(os.path.join(__dirname, ".node-red"))\
        .set_node_red_dir(os.path.join(__dirname, "node_red_dir"))\
        .set_admin_root("/node-red")\
        .set_port(1880)\
        .set_remote_access(False).build()

    # # init RED object directly
    # red = RED(
    #     os.path.join(__dirname, ".node-red"),
    #     os.path.join(__dirname, "node_red_dir"),
    #     "/node-red", 
    #     port = 1880
    # )

    # set editor theme
    red.editor_theme.palette.editable = False
    red.editor_theme.projects.enabled = False

    # add auths
    red.node_auths.append(
        Auth(username = "node-red-py", password = "p@ssword")
    )

    # register routes
    @route("/nodered-py-api/test", "get")
    def test_route(params:dict) -> dict:
        return { "state": "success" }
    
    @route("/nodered-py-api/test2", "get")
    def test_route2(params:dict) -> str:
        return """
<!DOCTYPE html>
<html>
    123412341234
</html>
"""

    @route("/nodered-py-api/test3", "post")
    def test_route3(datas:dict) -> dict:
        return { "state": "success" }
    
    red.static("/nodered-py-static", os.path.join(__dirname, "static"))

    app = TotalApp()
    red.register(app.test2, "test2")

    red.start(debug = True, start_browser = False)



# -*- coding: utf-8 -*-
import os, sys, subprocess, json, shutil
from glob import glob
from typing import List, Literal, Union
from types import MethodType
from .node import Node
from .route import Route, StaticRoute
from .theme import REDTheme
from .auth import AuthCollection
from ..__property__ import Property
from .. import __path__


class RED:
    """
    Node-RED manager class
    """
    registered_nodes:List[Node] = []
    registered_routes:List[Route] = []

    def __init__(self, user_dir:str, node_red_dir:str, admin_root:str, node_root:str, port:int, default_flow:str, remote_access:bool, default_categories:List[str]):
        """
        Set configs of Node-RED and setup

        Parameters
        ----------
        user_dir: str
            userDir of Node-RED settings
        node_red_dir: str
            directory for Node-RED starter
        admin_root: str
            httpAdminRoot of Node-RED settings
        node_root: str
            httpNodeRoot of Node-RED settings
        port: int
            port of Node-RED server
        default_flow: str
            flowFile of Node-RED settings
        remote_access: bool
            enable remote access of Node-RED or not
        default_categories: List[str]
            list of categories to show default
            (for detail information, see `Editor Configuration/paletteCategories` section of https://nodered.org/docs/user-guide/runtime/configuration)
        """
        self.user_dir, self.admin_root, self.node_root, self.port, self.default_flow, self.remote_access, self.__editor_theme, self.__node_auths =\
            user_dir, admin_root, node_root, port, default_flow, remote_access, REDTheme(), AuthCollection()
        self.__default_categories = default_categories
        self.__temp_dir, self.__node_dir =\
            os.path.join(__path__[0], ".temp"), os.path.join(__path__[0], ".nodejs")

        if not self.admin_root.startswith("/"):
            raise SyntaxError("`admin_root` must starts with '/'!")

        if not self.node_root.startswith("/"):
            raise SyntaxError("`node_root` must starts with '/'!")

        # check node.js exists
        try:
            subprocess.call(
                [ "npm.cmd" if sys.platform == "win32" else "npm", "--version" ],
                stdout = subprocess.DEVNULL,
                stderr = subprocess.STDOUT
            )
            self.__npm_path = "npm.cmd" if sys.platform == "win32" else "npm"
            self.__node_path = "node.exe" if sys.platform == "win32" else "node"
        except FileNotFoundError:
            if not os.path.exists(self.__node_dir):
                import platform, wget, zipfile, tarfile

                node_version = "18.16.1"
                if not os.path.exists(self.__temp_dir):
                    os.mkdir(self.__temp_dir)

                if sys.platform == "win32":
                    node_bin_zip = os.path.join(self.__temp_dir, "node.zip")
                    if platform.architecture()[0] == "32bit":
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-win-x86.zip", node_bin_zip)
                        zipfile.ZipFile(node_bin_zip).extractall(self.__temp_dir)
                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-win-x86"), self.__node_dir)
                    else:
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-win-x64.zip", node_bin_zip)
                        zipfile.ZipFile(node_bin_zip).extractall(self.__temp_dir)
                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-win-x64"), self.__node_dir)

                    self.__npm_path = os.path.join(self.__node_dir, "npm.exe")
                    self.__node_path = os.path.join(self.__node_dir, "node.exe")
                elif sys.platform == "darwin":
                    node_bin_zip = os.path.join(self.__temp_dir, "node.tar.gz")
                    if platform.processor() == "arm":
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-darwin-arm64.tar.gz", node_bin_zip)
                        with tarfile.open(node_bin_zip, "r", encoding = "utf-8") as nbzr:
                            nbzr.extractall(self.__temp_dir)

                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-darwin-arm64"), self.__node_dir)
                    else:
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-darwin-x64.tar.gz", node_bin_zip)
                        with tarfile.open(node_bin_zip, "r", encoding = "utf-8") as nbzr:
                            nbzr.extractall(self.__temp_dir)

                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-darwin-x64"), self.__node_dir)

                    self.__npm_path = os.path.join(self.__node_dir, "bin", "npm")
                    self.__node_path = os.path.join(self.__node_dir, "bin", "node")
                else:
                    node_bin_zip = os.path.join(self.__temp_dir, "node.tar.xz")
                    if platform.processor() == "arm":
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-linux-armv7l.tar.xz", node_bin_zip)
                        with tarfile.open(node_bin_zip, "r", encoding = "utf-8") as nbzr:
                            nbzr.extractall(self.__temp_dir)

                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-linux-armv7l"), self.__node_dir)
                    else:
                        wget.download(f"https://nodejs.org/dist/v{node_version}/node-v{node_version}-linux-x64.tar.xz", node_bin_zip)
                        with tarfile.open(node_bin_zip, "r", encoding = "utf-8") as nbzr:
                            nbzr.extractall(self.__temp_dir)

                        shutil.move(os.path.join(self.__temp_dir, f"node-v{node_version}-linux-x64"), self.__node_dir)

                shutil.rmtree(self.__temp_dir)

            self.__npm_path = os.path.join(self.__node_dir, "bin", "npm.cmd" if sys.platform == "win32" else "npm")
            self.__node_path = os.path.join(self.__node_dir, "bin", "node.exe" if sys.platform == "win32" else "node")
        
        # set node_red_dir
        if node_red_dir is None:
            self.node_red_dir = os.path.join(__path__[0], "node-red-starter")
        else:
            self.node_red_dir = node_red_dir = os.path.realpath(node_red_dir)
            if os.path.exists(node_red_dir):
                if not { "index.js", "package.json" }.issubset(set(os.listdir(node_red_dir))):
                    raise RuntimeError("Target `node_red_dir` is not Node-RED dir format!")

                shutil.copyfile(os.path.join(__path__[0], "node-red-starter", "route.js"), os.path.join(node_red_dir, "route.js"))
                shutil.copyfile(os.path.join(__path__[0], "node-red-starter", "index.js"), os.path.join(node_red_dir, "index.js"))
            else:
                os.mkdir(node_red_dir)
                shutil.copyfile(os.path.join(__path__[0], "node-red-starter", "route.js"), os.path.join(node_red_dir, "route.js"))
                shutil.copyfile(os.path.join(__path__[0], "node-red-starter", "index.js"), os.path.join(node_red_dir, "index.js"))
                shutil.copyfile(os.path.join(__path__[0], "node-red-starter", "package.json"), os.path.join(node_red_dir, "package.json"))

        self.__started_file = os.path.join(self.node_red_dir, "started")

        # setup Node-RED starter
        subprocess.call(
            [ self.__npm_path, "install" ],
            stdout = subprocess.DEVNULL,
            stderr = subprocess.STDOUT,
            cwd = self.node_red_dir
        )

    @property
    def editor_theme(self) -> REDTheme:
        """
        editorTheme of Node-RED server (for detail information, see https://github.com/node-red/node-red/wiki/Design:-Editor-Themes)
        """
        return self.__editor_theme
    
    @property
    def node_auths(self) -> AuthCollection:
        """
        adminAuth of Node-RED settings (for detail information, see `Editor & Admin API secutity` section of https://nodered.org/docs/user-guide/runtime/securing-node-red)
        """
        return self.__node_auths
    
    def __save_config(self, is_ready:bool):
        categories = []
        for node in RED.registered_nodes:
            if not node.category in categories:
                categories.append(node.category)

        for default_category in self.__default_categories:
            if not default_category in categories:
                categories.append(default_category)

        with open(os.path.join(self.node_red_dir, "config.json"), "w", encoding = "utf-8") as cfw:
            json.dump({
                "userDir": self.user_dir,
                "adminRoot": self.admin_root,
                "nodeRoot": self.node_root,
                "port": self.port,
                "defaultFlow": self.default_flow,
                "enableRemoteAccess": self.remote_access,
                "categories": categories,
                "editorTheme": self.editor_theme.to_dict(),
                "adminAuth": [] if is_ready else self.node_auths.to_list(),
                "cacheDir": os.path.join(self.user_dir, ".cache"),
                "routes": [
                    route.to_dict()
                    for route in RED.registered_routes
                ]
            }, cfw, indent = 4)
    
    def register(self, node_func:MethodType, name:str, category:str = "nodered_py", version:str = "1.0.0", description:str = "", author:str = "nodered.py", keywords:List[str] = [], icon:str = "function.png", properties:List[Property] = []):
        """
        Function to register Node function

        Parameters
        ----------
        node_func: MethodType, required
            Node function to register
        name: str, required
            name of Node to register
        category: str, default nodered_py
            category of Node
        version: str, default 1.0.0
            version of Node
        description: str, default ""
            description of Node
        author: str, default nodered.py
            author of Node
        keywords: List[str], default []
            extra keywords of Node
        icon: str, default function.png
            icon of Node (for detail information, see https://nodered.org/docs/creating-nodes/appearance)
        properties: List[noderedpy._property.Property]
            propertis of Node
        """
        RED.registered_nodes.append(
            Node(
                name, category,
                version, description, author, keywords,
                icon, properties, node_func
            )
        )

    def route(route_func:MethodType, url:str, method:Literal["get", "post"]):
        """
        Function to register route to Node-RED

        Parameters
        ----------
        route_func: MethodType, required
            route function to register
        url: str, required
            url of route point
        method: str, required
            method of route point
            options: get, post
        """
        RED.registered_routes.append(
            Route(url, method, route_func)
        )

    def static(self, url:str, path:os.PathLike):
        """
        Function to register route to Node-RED

        Parameters
        ----------
        url: str, required
            url of static point
        path: PathLike, required
            file path for static point
        """
        RED.registered_routes.append(
            StaticRoute(url, path)
        )

    # check input(node)
    def __check_node_input(self, node_input_file:os.PathLike, node_output_file:os.PathLike):
        # if node input exists
        if os.path.exists(node_input_file):
            # read input file
            while True:
                # if cannot read file or read during file writing, read file until can read
                try:
                    with open(node_input_file, "r", encoding = "utf-8") as ifr:
                        input_data = json.load(ifr)
                    
                    os.remove(node_input_file)
                    break
                except json.JSONDecodeError:
                    pass

            node = list(filter(lambda n: n.name == input_data["name"], RED.registered_nodes))[0]

            self.__write_node_output(
                node_output_file,
                node.run(input_data["props"], input_data["msg"])
            )

    # write output(node)
    def __write_node_output(self, output_file:str, res:dict):
        try:
            with open(output_file, "w", encoding = "utf-8") as ofw:
                json.dump(res, ofw, indent = 4)
        except:
            os.remove(output_file)
            res["req"]["body"] = {}
            self.__write_node_output(output_file, res)

    # check input(route)
    def __check_route_input(self, route_input_file:os.PathLike, route_output_file:os.PathLike):
        # if route input exists
        if os.path.exists(route_input_file):
            # read input file
            while True:
                # if cannot read file or read during file writing, read file until can read
                try:
                    with open(route_input_file, "r", encoding = "utf-8") as ifr:
                        input_data = json.load(ifr)
                    
                    os.remove(route_input_file)
                    break
                except json.JSONDecodeError:
                    pass

            route:Route = list(filter(lambda r: r.url == input_data["url"], RED.registered_routes))[0]
            self.__write_route_output(
                route_output_file,
                route.run(input_data["data"])
            )

    # write output(route)
    def __write_route_output(self, output_file:str, res:Union[str, dict]):
        with open(output_file, "w", encoding = "utf-8") as ofw:
            if isinstance(res, str):
                ofw.write(res)
            else:
                json.dump(res, ofw, indent = 4)


    def start(self, callback:MethodType = None, debug:bool = True, start_browser:bool = True):
        """
        Start Node-RED server

        Parameters
        ----------
        debug: bool, default True
            show outputs on console or not
        callback: MethodType, default None
            callback when Node-RED server started
        """

        # setup user_dir
        self.__start_for_ready()
        # kill if process listen on port
        self.stop()

        # set cache_dir
        self.__cache_dir = os.path.join(self.user_dir, ".cache")
        if os.path.exists(self.__cache_dir):
            shutil.rmtree(self.__cache_dir)

        os.mkdir(self.__cache_dir)

        # remove existing nodes
        for node_dir in glob(os.path.join(self.user_dir, "node_modules", "nodered-py-*")):
            shutil.rmtree(node_dir)

        # create custom nodes
        for node in RED.registered_nodes:
            node.create(self.user_dir, self.__cache_dir)

        if self.editor_theme.page.favicon is not None:
            favicon_file = os.path.join(self.node_red_dir, "favicon.ico")
            # convert png to ico if not ico file
            if not os.path.splitext(self.editor_theme.page.favicon)[-1] == ".ico":
                from PIL import Image

                Image.open(self.editor_theme.page.favicon).save(
                    favicon_file
                )
            # if ico, copy file
            else:
                shutil.copyfile(
                    self.editor_theme.page.favicon,
                    favicon_file
                )

            # self.editor_theme.page.favicon = favicon_file

        # save configs
        self.__save_config(False)

        # run Node-RED server
        subprocess.Popen([
            self.__node_path,
            "index.js"
        ], shell = False, stdout = sys.stdout if debug else subprocess.DEVNULL, stderr = subprocess.STDOUT, cwd = self.node_red_dir)

        while True:
            if os.path.exists(self.__started_file):
                if start_browser:
                    import webbrowser
                    webbrowser.open_new(f"http://127.0.0.1:{self.port}{self.admin_root}")

                if callback:
                    callback()

                break

        try:
            while True:
                self.__check_node_input(
                    os.path.join(self.__cache_dir, "node_input.json"),
                    os.path.join(self.__cache_dir, "node_output.json")
                )
                self.__check_route_input(
                    os.path.join(self.__cache_dir, "route_input.json"),
                    os.path.join(self.__cache_dir, "route_output.json")
                )
        except KeyboardInterrupt:
            self.stop()

    def __start_for_ready(self):
        """
        Start Node-RED for setup default userDir
        """
        # stop for safety
        self.stop()

        if os.path.exists(self.__started_file):
            os.remove(self.__started_file)

        # save configs
        self.__save_config(True)

        subprocess.Popen([
            self.__node_path,
            "index.js"
        ], shell = False, stdout = subprocess.DEVNULL, stderr = subprocess.STDOUT, cwd = self.node_red_dir)

        while True:
            if os.path.exists(self.__started_file):
                self.stop()
                break

        import time
        time.sleep(1)

    def stop(self):
        """
        Stop Node-RED server
        """
        import psutil, signal

        if os.path.exists(self.__started_file):
            os.remove(self.__started_file)

        killed = False
        for process in psutil.process_iter():
            try:
                for conns in process.connections(kind = "inet"):
                    if conns.laddr.port == self.port:
                        process.send_signal(signal.SIGTERM)
                        killed = True
                        break
            except ( psutil.AccessDenied, psutil.ZombieProcess ):
                pass

            if killed:
                break

class REDBuilder:
    def __init__(self):
        """
        Class to create RED using Builder-Pattern
        """
        self.__user_dir:str = None
        self.__node_red_dir:str = None
        self.__admin_root:str = "/node-red-py"
        self.__node_root:str = "/"
        self.__port:int = 1880
        self.__default_flow:str = "noderedpy.json"
        self.__remote_access:bool = True
        self.__default_categories:List[str] = [ "subflows", "common", "function", "network", "sequence", "parser", "storage" ]

    def set_user_dir(self, user_dir:str) -> "REDBuilder":
        """
        Function to set user_dir

        Parameters
        ----------
        user_dir: str
            userDir of Node-RED settings
        
        Return
        ------
        builder:REDBuilder
        """
        self.__user_dir = user_dir
        return self
    
    def set_node_red_dir(self, node_red_dir:str) -> "REDBuilder":
        """
        Function to set node_red_dir

        Parameters
        ----------
        node_red_dir: str
            directory for Node-RED starter
        
        Return
        ------
        builder:REDBuilder
        """
        self.__node_red_dir = node_red_dir
        return self
    
    def set_admin_root(self, admin_root:str) -> "REDBuilder":
        """
        Function to set admin_root

        Parameters
        ----------
        admin_root: str
            httpAdminRoot of Node-RED settings

        Return
        ------
        builder:REDBuilder
        """
        self.__admin_root = admin_root
        return self
    
    def set_node_root(self, node_root:str) -> "REDBuilder":
        """
        Function to set node_root

        Parameters
        ----------
        node_root: str
            httpNodeRoot of Node-RED settings
        
        Return
        ------
        builder:REDBuilder
        """
        self.__node_root = node_root
        return self
    
    def set_port(self, port:int) -> "REDBuilder":
        """
        Function to set port

        Parameters
        ----------
        port: int
            port of Node-RED server
        
        Return
        ------
        builder:REDBuilder
        """
        self.__port = port
        return self
    
    def set_default_flow(self, flow_file:str) -> "REDBuilder":
        """
        Function to set default_flow

        Parameters
        ----------
        flow_file: str
            flowFile of Node-RED settings

        Return
        ------
        builder:REDBuilder
        """
        self.__default_flow = flow_file
        return self

    def set_remote_access(self, remote_access:bool) -> "REDBuilder":
        """
        Function to set remote_access

        Parameters
        ----------
        remote_access: bool
            enable remote access of Node-RED or not
        
        Return
        ------
        builder:REDBuilder
        """
        self.__remote_access = remote_access
        return self
    
    def set_default_categories(self, default_categories:List[str]) -> "REDBuilder":
        """
        Function to set default_categories

        Parameters
        ----------
        default_categories: List[str]
            list of categories to show default
            (for detail information, see `Editor Configuration/paletteCategories` section of https://nodered.org/docs/user-guide/runtime/configuration)

        Return
        ------
        builder:REDBuilder
        """
        self.__default_categories = default_categories
        return self
    
    def build(self) -> RED:
        """
        Function to create RED from setups

        Return
        ------
        red:RED
        """

        if self.__user_dir is None:
            raise ValueError("`user_dir` must be set!")

        return RED(
            self.__user_dir, self.__node_red_dir,
            self.__admin_root, self.__node_root, self.__port, self.__default_flow,
            self.__remote_access, self.__default_categories
        )



# -*- coding: utf-8 -*-
import os, gc, json, traceback
from types import MethodType
from typing import List, Literal
from ..templates import package_json, node_html, node_js
from ..__property__ import Property
from .. import __path__



class Node:
    def __init__(self, name:str, category:str, version:str, description:str, author:str, keywords:List[str], icon:str, properties:List[Property], node_func:MethodType):
        # name of node cannot contain spaces
        if " " in name.strip():
            raise NameError("Node name cannot contain spaces!")
        
        # category of node cannot contain - or ,
        if "-" in category.strip() or "," in category.strip():
            raise NameError("Category cannot contain '-' or ','!")
        
        # remove default keyword in extra keywords
        self.keywords = [
            keyword
            for keyword in keywords
            if not keyword in ( "node-red", )
        ]

        self.name, self.category, self.version, self.description, self.author, self.icon, self.properties =\
            name, category, version, description, author, icon, properties

        self.__node_func = node_func

    def create(self, node_red_user_dir:str, node_red_user_cache_dir:str):
        self.__communicator = NodeCommunicator(os.path.join(node_red_user_cache_dir, "node_message.json"), self.name)
        node_dir = os.path.join(node_red_user_dir, "node_modules", self.name if self.name.startswith("nodered-py-") else f"nodered-py-{self.name}")
        os.makedirs(os.path.join(node_dir, "lib"))

        with open(os.path.join(node_dir, "package.json"), "w", encoding = "utf-8") as pjw:
            pjw.write(package_json(self))

        with open(os.path.join(node_dir, "lib", f"{self.name}.html"), "w", encoding = "utf-8") as nhw:
            nhw.write(node_html(self))

        with open(os.path.join(node_dir, "lib", f"{self.name}.js"), "w", encoding = "utf-8") as njw:
            njw.write(node_js(self, node_red_user_cache_dir))

    def run(self, props:dict, msg:dict) -> dict:
        gc.enable()

        print(f"\n{self.name} started\n===================================")
        try:
            resp = self.__node_func(self.__communicator, props, msg)
            print("============================= ended\n")

            resp.update({ "state": "success", "name": self.name })
            gc.collect()

            return resp
        except:
            return { "state": "fail", "name": self.name, "message": traceback.format_exc() }


class NodeCommunicator:
    def __init__(self, message_file:str, node_name:str):
        self.__message_file, self.__node_name = message_file, node_name

    def log(self, *args):
        with open(self.__message_file, "w", encoding = "utf-8") as mfw:
            json.dump({
                "name": self.__node_name,
                "log": args
            }, mfw, indent = 4)

    def warn(self, *args):
        with open(self.__message_file, "w", encoding = "utf-8") as mfw:
            json.dump({
                "name": self.__node_name,
                "warn": args
            }, mfw, indent = 4)

    def error(self, *args):
        with open(self.__message_file, "w", encoding = "utf-8") as mfw:
            json.dump({
                "name": self.__node_name,
                "error": args
            }, mfw, indent = 4)

    def status(self, fill:Literal["red", "green", "yellow", "blue", "grey"], shape:Literal["ring", "dot"], text:str):
        with open(self.__message_file, "w", encoding = "utf-8") as mfw:
            json.dump({
                "name": self.__node_name,
                "status": { "fill": fill, "shape": shape, "text": text }
            }, mfw, indent = 4)



# -*- coding: utf-8 -*-
import os, noderedpy
from ..__property__ import (
    PropertyDivider, Property,
    InputProperty, ListProperty,
    SpinnerProperty, CheckBoxProperty, ComboBoxProperty, CodeProperty,
    # FileProperty, TableProperty
)
from . import __path__


def package_json(node:"noderedpy.__nodered__.Node") -> str:
    with open(os.path.join(__path__[0], "template.json"), "r", encoding = "utf-8") as tr:
        tt = tr.read()

    tt = tt.replace(
        "{$node_name}", node.name
    ).replace(
        "{$node_version}", node.version
    ).replace(
        "{$node_description}", node.description
    ).replace(
        "{$node_author}", node.author
    )
    
    if node.keywords == []:
        tt = tt.replace(
            "{$node_keywords}", ""
        )
    else:
        tt = tt.replace(
            "{$node_keywords}", ", " + ", ".join([ f'"{keyword}"' for keyword in node.keywords ])
        )

    return tt

def node_html(node:"noderedpy.__nodered__.Node") -> str:
    properties_html, properties_js, properties_js_prepare, properties_js_cancel, properties_js_save = [], [], [], [], []

    for property in node.properties:
        if isinstance(property, InputProperty):
            properties_html.append(f"""
    <div class="form-row" style="margin-bottom:0px;">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
    </div>
    <div class="form-row">
        <input type="text" id="node-input-{property.var_name}" style="width:100%;">
    </div>
""")
            default_value = f'"{property.default}"' if isinstance(property.default, str) else str(property.default)
        elif isinstance(property, ListProperty):
            properties_html.append(f"""
    <div class="form-row" style="margin-bottom:0px;">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
    </div>
    <div class="form-row node-input-{property.var_name}-container-row">
        <ol id="node-input-{property.var_name}-container" style="height:{property.height}px;"></ol>
    </div>
""")
            properties_js_prepare.append('''
            $("#node-input-''' + property.var_name + '''-container").editableList({
                addItem: (container, idx, opt) => {
                    opt.idx = idx;
                    opt.value = opt.value ?? "";

                    container.css({ overflow: "hidden", display: "flex", "align-items": "center" });
                    $("<input>", { class: "input-list-item", type: "text", style:"flex:1;" }).val(opt.value).appendTo(container);
                },
                removable: true
            });

            for (var item of node["''' + property.var_name + '''"]) {
                $("#node-input-''' + property.var_name + '''-container").editableList("addItem", { value: item });
            }
''')
            properties_js_save.append('''
            node["''' + property.var_name + '''"] = [];
            $("#node-input-''' + property.var_name + '''-container").editableList("items").each((_, item) => {
                node["''' + property.var_name + '''"].push(item.find("input.input-list-item").val());
            });
''')
            default_value = str(property.default)
        elif isinstance(property, CodeProperty):
            properties_html.append(f"""
    <div class="form-row" style="margin-bottom:0px;">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
    </div>
    <div class="form-row node-text-editor-row">
        <div style="height:{property.height}px;" class="node-text-editor" id="node-input-{property.var_name}"></div>
    </div>
""")
            if property.language:
                properties_js_prepare.append('''
            node["''' + property.var_name + '''Editor"] = RED.editor.createEditor({
                id: "node-input-''' + property.var_name + '''",
                mode: "ace/mode/''' + property.language + '''",
                value: node["''' + property.var_name + '''"],
                focus: true
            });
''')
            else:
                properties_js_prepare.append('''
            node["''' + property.var_name + '''Editor"] = RED.editor.createEditor({
                id: "node-input-''' + property.var_name + '''",
                value: node["''' + property.var_name + '''"],
                focus: true
            });
''')
            properties_js_cancel.append(f"""
            node["{property.var_name}Editor"].destroy();
            delete node["{property.var_name}Editor"];
""")
            properties_js_save.append(f"""
            $("#node-input-{property.var_name}").val(node["{property.var_name}Editor"].getValue().trim());
            node["{property.var_name}Editor"].destroy();
            delete node["{property.var_name}Editor"];
""")
            default_value = f"`{property.default}`"
        elif isinstance(property, SpinnerProperty):
            properties_html.append(f"""
    <div class="form-row" style="margin-bottom:0px;">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
    </div>
    <div class="form-row">
        <input type="text" id="node-input-{property.var_name}" style="width:calc(100% - 22px);">
    </div>
""")
            spinner_configs = []
            if property.min:
                spinner_configs.append(f"                min: {property.min},")
            if property.max:
                spinner_configs.append(f"                max: {property.max},")
            if property.step:
                spinner_configs.append(f"                step: {property.step}")

            properties_js_prepare.append('''
            $("#node-input-''' + property.var_name + '''").spinner({
                ''' + "\n".join(spinner_configs) + '''
            });
''')
            default_value = str(property.default)
        elif isinstance(property, CheckBoxProperty):
            properties_html.append(f"""
    <div class="form-row">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
        <input type="checkbox" style="margin-left:10px;width:15px;height:15px;margin-bottom:5px;" id="node-input-{property.var_name}">
    </div>
""")
            default_value = "true" if property.default else "false"
        elif isinstance(property, ComboBoxProperty):
            options = "\n".join([
                f"""
            <option value="{item}">{item}</option>
"""
                for item in property.items
            ])
            properties_html.append(f"""
    <div class="form-row" style="margin-bottom:0px;">
        <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
    </div>
    <div class="form-row">
        <select id="node-input-{property.var_name}Select" style="width:100%;">
{options}
        </select>
        <input type="text" id="node-input-{property.var_name}" style="display:none;">
    </div>
""")
            properties_js_prepare.append('''
            $("#node-input-''' + property.var_name + '''Select").val($("#node-input-''' + property.var_name + '''").val());
''')
            properties_js_save.append('''
            $("#node-input-''' + property.var_name + '''").val($("#node-input-''' + property.var_name + '''Select").val());
''')
            default_value = f'"{property.default}"' if isinstance(property.default, str) else str(property.default)
#         elif isinstance(property, FileProperty):
#             properties_html.append(f"""
#     <div class="form-row" style="display:flex;flex-flow:row;align-items:center;">
#         <label style="width:auto;margin-bottom:0px;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
#         <input type="file" id="node-input-{property.name}" style="flex:1;">
#     </div>
# """)
#             default_value = f'"{property.default}"'
#         elif isinstance(property, TableProperty):
#             theader = "\n".join([ f"                <th>{column}</th>" for column in property.default.columns ])
#             properties_html.append(f"""
#     <div class="form-row" style="margin-bottom:0px;">
#         <label style="width:auto;"><i class="{property.display_icon}"></i> <span>{property.display_name}</span></label>
#     </div>
#     <div class="form-row" style="display:flex;flex-flow:column;">
#         <table>
#             <thead>
# {theader}
#             </thead>
#             <tbody>
#             </tbody>
#         </table>
#     </div>
# """)
        else:
            if isinstance(property, PropertyDivider):
                properties_html.append("""
    <hr>
""")
            default_value = None

        if default_value:
            required = 'true' if property.required else 'false'
            properties_js.append('            "' + property.var_name + '": { value: ' + default_value + ', required: ' + required + ' }')

    with open(os.path.join(__path__[0], "template.html"), "r", encoding = "utf-8") as tr:
        tt = tr.read()

    return tt.replace(
        "{$node_name}", node.name
    ).replace(
        "{$node_category}", node.category
    ).replace(
        "{$node_icon}", f'"{node.icon}"' if node.icon else "null"
    ).replace(
        "{$properties_html}", "".join(properties_html)
    ).replace(
        "{$properties_js}", ",\n".join(properties_js)
    ).replace(
        "{$properties_js_prepare}", "\n".join(properties_js_prepare)
    ).replace(
        "{$properties_js_cancel}", "\n".join(properties_js_cancel)
    ).replace(
        "{$properties_js_save}", "\n".join(properties_js_save)
    )

def node_js(node:"noderedpy.__nodered__.Node", cache_dir:str) -> str:
    with open(os.path.join(__path__[0], "template.js"), "r", encoding = "utf-8") as tr:
        tt = tr.read()

    return tt.replace(
        "{$node_name}", node.name
    ).replace(
        "{$node_properties}", str([ property.var_name for property in node.properties if isinstance(property, Property)])
    ).replace(
        "{$cache_dir}", cache_dir
    )



# -*- coding: utf-8 -*-
import os
from dataclasses import dataclass, field
from .. import __path__


@dataclass
class PageTheme:
    title:str = field(default = "Node-RED.py")
    favicon:str = field(default = os.path.join(__path__[0], "assets", "python-logo.png"))

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "favicon": self.favicon
        }

@dataclass
class HeaderTheme:
    title:str = field(default = "Node-RED.py")
    image:str = field(default = None)

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "image": self.image
        }

@dataclass
class MenuTheme:
    menu_item_palette:bool = False

    def to_dict(self) -> dict:
        return {
            "menu-item-palette": self.menu_item_palette
        }

@dataclass
class PaletteTheme:
    editable:bool = True

    def to_dict(self) -> dict:
        return {
            "editable": self.editable
        }

@dataclass
class ProjectsTheme:
    enabled:bool = True

    def to_dict(self) -> dict:
        return {
            "enabled": self.enabled
        }

class REDTheme:
    def __init__(self):
        self.__page_theme, self.__header_theme, self.__menu_theme, self.__palette_theme, self.__projects_theme =\
            PageTheme(), HeaderTheme(), MenuTheme(), PaletteTheme(), ProjectsTheme()
        self.__tours, self.__user_menu = False, True

    @property
    def page(self) -> PageTheme:
        return self.__page_theme
    
    @property
    def header(self) -> HeaderTheme:
        return self.__header_theme
    
    @property
    def menu(self) -> MenuTheme:
        return self.__menu_theme
    
    @property
    def tours(self) -> bool:
        return self.__tours
    
    @tours.setter
    def tours(self, value:bool):
        self.__tours = value

    @property
    def user_menu(self) -> bool:
        return self.__user_menu
    
    @user_menu.setter
    def user_menu(self, value:bool):
        self.__user_menu = value
    
    @property
    def palette(self) -> PaletteTheme:
        return self.__palette_theme
    
    @property
    def projects(self) -> ProjectsTheme:
        return self.__projects_theme

    def to_dict(self) -> dict:
        return {
            "page": self.page.to_dict(),
            "header": self.header.to_dict(),
            "menu": self.menu.to_dict(),
            "tours": self.tours,
            "userMenu": self.user_menu,
            "palette": self.palette.to_dict(),
            "projects": self.projects.to_dict()
        }



# -*- coding: utf-8 -*-
import json
from typing import Any, Union, List



class Property:
    def __init__(self, name:str, default:Any = None, required:bool = False, display_name:str = None, display_icon:str = None):
        self.name, self.default, self.required, self.display_icon =\
            name, default, required, display_icon
        self.display_name = " ".join([ item.strip().capitalize() for item in name.split("_") ]) if display_name is None else display_name
    
    @property
    def var_name(self) -> str:
        return f"np-var_{self.name}"

class PropertyDivider:
    pass

class CodeProperty(Property):
    def __init__(self, name:str, default:str = "", language:str = None, height:int = 250, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to edit code

        name: str, required
            name of CodeProperty
        default: str, default ""
            default value of CodeProperty
        language: str, default None
            language mode("ace/mode/") of Node-RED Editor
        height: int, default 250
            height to display in Node-RED edit dialog
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if not isinstance(default, str):
            raise TypeError("CodeProperty can accept types: [ 'dict', 'json string' ]")
        
        super().__init__(name, default, required, display_name, display_icon if display_icon else "fa fa-code")
        self.language, self.height =\
            language, height

class InputProperty(Property):
    def __init__(self, name:str, default:Union[int, float, str] = None, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to change value

        name: str, required
            name of InputProperty
        default: Union[int, float, str], default ""
            default value of InputProperty
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if default is None:
            if isinstance(default, int):
                default = 0
            elif isinstance(default, float):
                default = 0.0
            else:
                default = ""
        elif not isinstance(default, (int, float, str)):
            raise TypeError("InputProperty can accept types: [ 'int', 'float', 'str' ]")

        super().__init__(name, default, required, display_name, display_icon)

        if display_icon is None:
            if isinstance(default, (int, float)):
                self.display_icon = "fa fa-sort-numeric-asc"
            else:
                self.display_icon = "fa fa-font"
        else:
            self.display_icon = display_icon

class ListProperty(Property):
    def __init__(self, name:str, default:list = [], height:int = 250, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to handle list

        name: str, required
            name of ListProperty
        default: list, default []
            default value of ListProperty
        height: int, default 250
            height to display in Node-RED edit dialog
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if not isinstance(default, list):
            raise TypeError("ListProperty can accept type: 'list'")

        super().__init__(name, default, required, display_name, display_icon if display_icon else "fa fa-list-ul")
        self.height = height

class DictProperty(CodeProperty):
    def __init__(self, name:str, default:Union[dict, str] = {}, height:int = 250, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to handle dict

        name: str, required
            name of Dictroperty
        default: Union[dict, str], default {}
            default value of DictProperty
        height: int, default 250
            height to display in Node-RED edit dialog
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if not isinstance(default, ( dict, str )):
            raise TypeError("DictProperty can accept types: [ 'dict', 'json string' ]")
        
        if isinstance(default, str):
            if not default.strip().startswith("{") or not default.strip().endswith("}"):
                raise ValueError("DictProperty value must be 'dict' or 'json string'!")

            try:
                json.loads(default)
            except:
                raise ValueError("DictProperty value must be 'dict' or 'json string'!")

        if isinstance(default, dict):
            default = json.dumps(default, indent = 4)

        super().__init__(name, default, "json", height, required, display_name, display_icon if display_icon else "fa fa-file-code-o")

class SpinnerProperty(Property):
    def __init__(self, name:str, default:float = 0, step:float = None, min:float = None, max:float = None, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to handle float with spinner

        name: str, required
            name of SpinnerProperty
        default: float, default 0
            default value of SpinnerProperty
        step: float, default None
            step for spinner
        min: float, default None
            min value for spinner
        max: float, default None
            max value for spinner
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if not isinstance(default, (int, float)):
            raise TypeError("SpinnerProperty can accept types: [ 'int', 'float' ]")

        super().__init__(name, default, required, display_name, display_icon if display_icon else "fa fa-random")
        self.step, self.min, self.max =\
            step, min, max

class CheckBoxProperty(Property):
    def __init__(self, name:str, default:bool = False, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to handle checked state

        name: str, required
            name of CheckBoxProperty
        default: bool, default False
            default value of CheckBoxProperty
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        super().__init__(name, default, required, display_name, display_icon if display_icon else "fa fa-check")

class ComboBoxProperty(Property):
    def __init__(self, name:str, items:List[Any], default:Any = None, required:bool = False, display_name:str = None, display_icon:str = None):
        """
        Property to select value from lists

        name: str, required
            name of ComboBoxProperty
        items: List[Any], required
            items of ComboBoxProperty
        default: Any, default None
            default value of ComboBoxProperty
        required: bool, default False
            set required or not
        display_name: str, default None
            name to display in Node-RED edit dialog
        display_icon: str, default None
            icon to display in Node-RED edit dialog (for available icons, see https://fontawesome.com/v4/icons/)
        """
        if not isinstance(items, list):
            raise TypeError("items of ComboBoxProperty must be type: 'list'")
        
        if default is None:
            if len(items) > 0:
                default = items[0]
            else:
                default = ""
        
        super().__init__(name, default, required, display_name, display_icon if display_icon else "fa fa-filter")
        self.items = items

# class FileProperty(Property):
#     def __init__(self, name:str, default:str = None, required:bool = False, display_icon:str = None):
#         super().__init__(name, default if default else "", required, display_icon if display_icon else "fa fa-file-text-o")

# class TableProperty(Property):
#     def __init__(self, name:str, default:pd.DataFrame = None, required:bool = False, display_icon:str = None):
#         super().__init__(name, default if default else pd.DataFrame(), required, display_icon if display_icon else "fa fa-table")



# -*- coding: utf-8 -*-
"""
make python function to Node-RED node
"""

from .__nodered__.red import RED, REDBuilder
from .__nodered__.node import NodeCommunicator as Node
from .__nodered__.auth import Auth
from .__property__ import (
    PropertyDivider,
    InputProperty, ListProperty, DictProperty, CodeProperty,
    SpinnerProperty, CheckBoxProperty, ComboBoxProperty
)

__version__ = "0.2.12"
__author__ = "oyajiDev"
__email__ = "this.dev.somehit@gmail.com"

__all__ = [
    "RED", "REDBuilder", "Auth", "Node",
    "PropertyDivider", "InputProperty", "ListProperty", "DictProperty", "CodeProperty",
    "SpinnerProperty", "CheckBoxProperty", "ComboBoxProperty"
]
